#ifndef MSP_GEOMETRY_HYPERBOX_H_
#define MSP_GEOMETRY_HYPERBOX_H_

#include <algorithm>
#include <cmath>
#include <msp/linal/vector.h>
#include "ray.h"
#include "shape.h"
#include "surfacepoint.h"

namespace Msp {
namespace Geometry {

/**
A shape bounded by planar faces at right angles to each other.  Two- and three-
dimensional cases are Rectangle and Box, respectively.
*/
template<typename T, unsigned D>
class HyperBox: public Shape<T, D>
{
private:
	LinAl::Vector<T, D> dimensions;

public:
	HyperBox();
	explicit HyperBox(const LinAl::Vector<T, D> &);

	virtual HyperBox *clone() const;

	const LinAl::Vector<T, D> &get_dimensions() const { return dimensions; }
	T get_dimension(unsigned) const;

	virtual HyperBox<T, D> get_axis_aligned_bounding_box() const { return *this; }
	virtual bool contains(const LinAl::Vector<T, D> &) const;
	virtual bool check_intersection(const Ray<T, D> &) const;
	virtual unsigned get_max_ray_intersections() const { return 2; }
	virtual unsigned get_intersections(const Ray<T, D> &, SurfacePoint<T, D> *, unsigned) const;
};

template<typename T, unsigned D>
inline HyperBox<T, D>::HyperBox()
{
	for(unsigned i=0; i<D; ++i)
		dimensions[i] = 1;
}

template<typename T, unsigned D>
inline HyperBox<T, D>::HyperBox(const LinAl::Vector<T, D> &d):
	dimensions(d)
{ }

template<typename T, unsigned D>
inline HyperBox<T, D> *HyperBox<T, D>::clone() const
{
	return new HyperBox<T, D>(dimensions);
}

template<typename T, unsigned D>
inline T HyperBox<T, D>::get_dimension(unsigned i) const
{
	return dimensions[i];
}

template<typename T, unsigned D>
inline bool HyperBox<T, D>::contains(const LinAl::Vector<T, D> &point) const
{
	using std::abs;

	for(unsigned i=0; i<D; ++i)
		if(abs(point[i])>dimensions[i]/2)
			return false;
	return true;
}

template<typename T, unsigned D>
inline bool HyperBox<T, D>::check_intersection(const Ray<T, D> &ray) const
{
	return get_intersections(ray, 0, 1);
}

template<typename T, unsigned D>
inline unsigned HyperBox<T, D>::get_intersections(const Ray<T, D> &ray, SurfacePoint<T, D> *points, unsigned size) const
{
	using std::abs;

	LinAl::Vector<T, D> half_dim = dimensions/T(2);
	unsigned n = 0;
	for(unsigned i=0; i<D; ++i)
	{
		if(!ray.get_direction()[i])
			continue;

		for(int j=-1; j<=1; j+=2)
		{
			T x = (T(j)*half_dim[i]-ray.get_start()[i])/ray.get_direction()[i];
			if(!ray.check_limits(x))
				continue;

			LinAl::Vector<T, D> p = ray.get_start()+ray.get_direction()*x;

			bool inside = true;
			for(unsigned k=0; (inside && k<D); ++k)
				inside = (k==i || abs(p[k])<=half_dim[k]);

			if(inside && n<size)
			{
				if(points)
				{
					points[n].position = p;
					points[n].normal = LinAl::Vector<T, D>();
					points[n].normal[i] = j;
					points[n].distance = x;

					if(n==1 && x<points[0].distance)
						std::swap(points[0], points[1]);
				}

				++n;
				if(n==size || n==2)
					return n;
			}
		}
	}

	return n;
}

} // namespace Geometry
} // namespace Msp

#endif
